//--------------------------------------------------------------------------------------
// File: dxhelper.h
//
// (FreeBSD License)
// 
// Copyright 2010 Andreas Kirsch. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice, this list of
// conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list
// of conditions and the following disclaimer in the documentation and/or other materials
// provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY ANDREAS KIRSCH ``AS IS'' AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ANDREAS KIRSCH OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
// ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// The views and conclusions contained in the software and documentation are those of the
// authors and should not be interpreted as representing official policies, either expressed
// or implied, of Andreas Kirsch.
//--------------------------------------------------------------------------------------
#pragma once
#include <D3D10.h>

void createBuffer( ID3D10Device *device, ID3D10Buffer **buffer, unsigned size, D3D10_USAGE usage, D3D10_BIND_FLAG bindFlags, void *data = NULL, D3D10_CPU_ACCESS_FLAG cpuFlag = (D3D10_CPU_ACCESS_FLAG) 0 );

static const D3DXVECTOR3 D3D10CubeMapFaceDirections[] = {
	D3DXVECTOR3(  1.0f, 0.0f, 0.0f ), // X+
	D3DXVECTOR3( -1.0f, 0.0f, 0.0f ), // X-
	D3DXVECTOR3( 0.0f,  1.0f, 0.0f ), // Y+
	D3DXVECTOR3( 0.0f, -1.0f, 0.0f ), // Y-
	D3DXVECTOR3( 0.0f, 0.0f,  1.0f ), // Z+
	D3DXVECTOR3( 0.0f, 0.0f, -1.0f ), // Z-
};

enum D3D10CubeMapFace {
	D3D10_CMF_POSITIVE_X,
	D3D10_CMF_NEGATIVE_X,
	D3D10_CMF_POSITIVE_Y,
	D3D10_CMF_NEGATIVE_Y,
	D3D10_CMF_POSITIVE_Z,
	D3D10_CMF_NEGATIVE_Z,
	D3D10_CMF_COUNT
};

// normal LHS to cubemap face
inline void D3D10MatrixLoadCubeMapFaceOrientation( D3DXMATRIX *pOut, D3D10CubeMapFace face ) {
	switch( face ) {
	case D3D10_CMF_POSITIVE_X:
		D3DXMatrixRotationY( pOut, D3DX_PI / 2 );
		break;
	case D3D10_CMF_NEGATIVE_X:
		D3DXMatrixRotationY( pOut, -D3DX_PI / 2 );
		break;
	case D3D10_CMF_POSITIVE_Y:
		D3DXMatrixRotationX( pOut, D3DX_PI / 2 );
		break;
	case D3D10_CMF_NEGATIVE_Y:
		D3DXMatrixRotationX( pOut, -D3DX_PI / 2 );
		break;
	case D3D10_CMF_POSITIVE_Z:
		D3DXMatrixIdentity( pOut );
		break;
	case D3D10_CMF_NEGATIVE_Z:
		D3DXMatrixRotationY( pOut, D3DX_PI );
		break;
	default:
		assert( false );
		D3DXMatrixIdentity( pOut );
		break;
	}
}

struct DXGI_SAMPLE_DESC_CPP : DXGI_SAMPLE_DESC {
	DXGI_SAMPLE_DESC_CPP() {
		Count = 1;
		Quality = 0;
	}
};

/*
struct D3D10_SUBRESOURCE_DATA_CPP
{
	const void *pSysMem;
	UINT SysMemPitch;
	UINT SysMemSlicePitch;
};*/

struct D3D10SubResourceData {
	D3D10_SUBRESOURCE_DATA desc;

	D3D10SubResourceData( const void *data = NULL ) {
		desc.pSysMem = data;
		desc.SysMemPitch = 0;
		desc.SysMemSlicePitch = 0;
	}

	operator D3D10_SUBRESOURCE_DATA*() {
		return &desc;
	}

	D3D10SubResourceData &sysMemPitch( UINT value ) {
		desc.SysMemPitch = value;
		return *this;
	}
};


struct D3D10_BIND_FLAG_CPP {
	UINT BindFlags;

	D3D10_BIND_FLAG_CPP() : BindFlags( 0 ) {}

	D3D10_BIND_FLAG_CPP& useAsShaderResource() { BindFlags |= D3D10_BIND_SHADER_RESOURCE; return *this; }
	D3D10_BIND_FLAG_CPP& useAsRenderTarget() { BindFlags |= D3D10_BIND_RENDER_TARGET; return *this; }
	D3D10_BIND_FLAG_CPP& useAsDepthStencil() { BindFlags |= D3D10_BIND_DEPTH_STENCIL; return *this; }
	D3D10_BIND_FLAG_CPP& useAsVertexBuffer() { BindFlags |= D3D10_BIND_VERTEX_BUFFER; return *this; }

	operator UINT() { return BindFlags; }
};

struct D3D10_TEXTURE2D_DESC_CPP : D3D10_TEXTURE2D_DESC {
	D3D10_TEXTURE2D_DESC_CPP( DXGI_FORMAT format, D3D10_USAGE usage, UINT width = 0, UINT height = 0 ) {
		Width = width;
		Height = height;
		Format = format;
		Usage = usage;
		MipLevels = 0;
		ArraySize = 1;
		SampleDesc = DXGI_SAMPLE_DESC_CPP();
		BindFlags = 0;
		CPUAccessFlags = 0;
		MiscFlags = 0;
	}

	void setSize( UINT width, UINT height ) {
		Width = width;
		Height = height;
	}

	void noMipMapLevels() {
		MipLevels = 1;
	}

	void arraySize( int size ) {
		ArraySize = size;
	}

	D3D10_BIND_FLAG_CPP & bindFlags() {
		return *(D3D10_BIND_FLAG_CPP*)&BindFlags;
	}

	void allowCPURead() {
		CPUAccessFlags |= D3D10_CPU_ACCESS_READ;
	}

	void allowCPUWrite() {
		CPUAccessFlags |= D3D10_CPU_ACCESS_WRITE;
	}

	void asTextureCube() {
		MiscFlags |= D3D10_RESOURCE_MISC_TEXTURECUBE;
	}

	void canGenerateMipMaps() {
		MiscFlags |= D3D10_RESOURCE_MISC_GENERATE_MIPS;
	}
};

struct D3D10_TEXTURE3D_DESC_CPP : D3D10_TEXTURE3D_DESC {
	D3D10_TEXTURE3D_DESC_CPP( DXGI_FORMAT format, D3D10_USAGE usage, UINT width = 0, UINT height = 0, UINT depth = 0 ) {
		Width = width;
		Height = height;
		Depth = depth;
		Format = format;
		Usage = usage;
		MipLevels = 0;
		BindFlags = 0;
		CPUAccessFlags = 0;
		MiscFlags = 0;
	}

	void setSize( UINT width, UINT height, UINT depth ) {
		Width = width;
		Height = height;
		Depth = depth;
	}

	void noMipMapLevels() {
		MipLevels = 1;
	}

	D3D10_BIND_FLAG_CPP & bindFlags() {
		return *(D3D10_BIND_FLAG_CPP*)&BindFlags;
	}

	void allowCPURead() {
		CPUAccessFlags |= D3D10_CPU_ACCESS_READ;
	}

	void allowCPUWrite() {
		CPUAccessFlags |= D3D10_CPU_ACCESS_WRITE;
	}

	void canGenerateMipMaps() {
		MiscFlags |= D3D10_RESOURCE_MISC_GENERATE_MIPS;
	}
};

struct D3D10Texture2DDesc {
	D3D10_TEXTURE2D_DESC_CPP desc;

	D3D10Texture2DDesc( DXGI_FORMAT Format, D3D10_USAGE Usage, UINT Width, UINT Height )
		: desc( Format, Usage, Width, Height ) {}

	D3D10Texture2DDesc & noMipMapLevels() { desc.noMipMapLevels(); return *this; }
	D3D10Texture2DDesc & arraySize( int size ) { desc.arraySize( size ); return *this; }
	D3D10Texture2DDesc & allowCPURead() { desc.allowCPURead(); return *this; }
	D3D10Texture2DDesc & allowCPUWrite() { desc.allowCPUWrite(); return *this; }

	D3D10Texture2DDesc & asTextureCube() { desc.asTextureCube(); return *this; }
	D3D10Texture2DDesc & canGenerateMipMaps()  { desc.canGenerateMipMaps();	return *this; }

	D3D10Texture2DDesc & bindFlags( const D3D10_BIND_FLAG_CPP &flags ) {
		desc.bindFlags() = flags;
		return *this;
	}

	operator D3D10_TEXTURE2D_DESC*() {
		return &desc;
	}
};

struct D3D10Texture3DDesc {
	D3D10_TEXTURE3D_DESC_CPP desc;

	D3D10Texture3DDesc( DXGI_FORMAT Format, D3D10_USAGE Usage, UINT Width, UINT Height, UINT Depth )
		: desc( Format, Usage, Width, Height, Depth ) {}

	D3D10Texture3DDesc & noMipMapLevels() { desc.noMipMapLevels(); return *this; }
	D3D10Texture3DDesc & allowCPURead() { desc.allowCPURead(); return *this; }
	D3D10Texture3DDesc & allowCPUWrite() { desc.allowCPUWrite(); return *this; }

	D3D10Texture3DDesc & canGenerateMipMaps()  { desc.canGenerateMipMaps();	return *this; }

	D3D10Texture3DDesc & bindFlags( const D3D10_BIND_FLAG_CPP &flags ) {
		desc.bindFlags() = flags;
		return *this;
	}

	operator D3D10_TEXTURE3D_DESC*() {
		return &desc;
	}
};

/*
struct D3D10_TEX2D_SRV_CPP : D3D10_TEX2D_SRV {
	D3D10_TEX2D_SRV_CPP() {
		MipLevels = 1;
		MostDetailedMip = 0;
	}

	D3D10_TEX2D_SRV_CPP &mipLevels( UINT levels ) { MipLevels = levels; return *this; }
	D3D10_TEX2D_SRV_CPP &mostDetailedMip( UINT level ) { MostDetailedMip = level; return *this; }
};

struct D3D10_SHADER_RESOURCE_VIEW_DESC_CPP : D3D10_SHADER_RESOURCE_VIEW_DESC {
	D3D10_SHADER_RESOURCE_VIEW_DESC_CPP( DXGI_FORMAT format ) {
		Format = format;
		ViewDimension = D3D10_SRV_DIMENSION_UNKNOWN;
	}

/ *
	void setTexture2D( UINT MostDetailedMip, UINT MipLevels ) {
		ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2D;
		Texture2D.MostDetailedMip = MostDetailedMip;
		Texture2D.MipLevels = MipLevels;
	}* /

	D3D10_SHADER_RESOURCE_VIEW_DESC_CPP & texture2D( const D3D10_TEX2D_SRV_CPP &desc = D3D10_TEX2D_SRV_CPP() ) {
		ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2D;
		Texture2D = desc;
		return *this;
	}
};*/

// NOTE:: all the *_VIEW_DESC structs are not binary compatible with each other (because of the DIMENSION enums!) [5/13/2010 Andreas Kirsch]
struct D3D10Texture2DShaderResourceViewDesc : D3D10_SHADER_RESOURCE_VIEW_DESC {
	D3D10Texture2DShaderResourceViewDesc( DXGI_FORMAT format ) {
		Format = format;
		ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2D;
		Texture2D.MipLevels = 1;
		Texture2D.MostDetailedMip = 0;
	}

	D3D10Texture2DShaderResourceViewDesc &mipLevels( UINT levels ) { Texture2D.MipLevels = levels; return *this; }
	D3D10Texture2DShaderResourceViewDesc &mostDetailedMip( UINT level ) { Texture2D.MostDetailedMip = level; return *this; }

	operator D3D10_SHADER_RESOURCE_VIEW_DESC*() {
		return this;
	}
};

struct D3D10Texture2DArrayShaderResourceViewDesc : D3D10_SHADER_RESOURCE_VIEW_DESC {
	D3D10Texture2DArrayShaderResourceViewDesc( DXGI_FORMAT format ) {
		Format = format;
		ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2DARRAY;
		Texture2DArray.MipLevels = 1;
		Texture2DArray.MostDetailedMip = 0;
		Texture2DArray.FirstArraySlice = 0;
		Texture2DArray.ArraySize = 1;
	}

	D3D10Texture2DArrayShaderResourceViewDesc &mipLevels( UINT levels ) { Texture2DArray.MipLevels = levels; return *this; }
	D3D10Texture2DArrayShaderResourceViewDesc &mostDetailedMip( UINT level ) { Texture2DArray.MostDetailedMip = level; return *this; }
	D3D10Texture2DArrayShaderResourceViewDesc &firstArraySlice( int slice ) { Texture2DArray.FirstArraySlice = slice; return *this; }
	D3D10Texture2DArrayShaderResourceViewDesc &arraySize( int size ) { Texture2DArray.ArraySize = size; return *this; }

	operator D3D10_SHADER_RESOURCE_VIEW_DESC*() {
		return this;
	}
};


struct D3D10TextureCubeMapShaderResourceViewDesc : D3D10_SHADER_RESOURCE_VIEW_DESC {
	D3D10TextureCubeMapShaderResourceViewDesc( DXGI_FORMAT format ) {
		Format = format;
		ViewDimension = D3D10_SRV_DIMENSION_TEXTURECUBE;
		TextureCube.MipLevels = 1;
		TextureCube.MostDetailedMip = 0;
	}

	D3D10TextureCubeMapShaderResourceViewDesc &mipLevels( UINT levels ) { TextureCube.MipLevels = levels; return *this; }
	D3D10TextureCubeMapShaderResourceViewDesc &mostDetailedMip( UINT level ) { TextureCube.MostDetailedMip = level; return *this; }

	operator D3D10_SHADER_RESOURCE_VIEW_DESC*() {
		return this;
	}
};

struct D3D10Texture3DShaderResourceViewDesc : D3D10_SHADER_RESOURCE_VIEW_DESC {
	D3D10Texture3DShaderResourceViewDesc( DXGI_FORMAT format ) {
		Format = format;
		ViewDimension = D3D10_SRV_DIMENSION_TEXTURE3D;
		TextureCube.MipLevels = 1;
		TextureCube.MostDetailedMip = 0;
	}

	D3D10Texture3DShaderResourceViewDesc &mipLevels( UINT levels ) { Texture3D.MipLevels = levels; return *this; }
	D3D10Texture3DShaderResourceViewDesc &mostDetailedMip( UINT level ) { Texture3D.MostDetailedMip = level; return *this; }

	operator D3D10_SHADER_RESOURCE_VIEW_DESC*() {
		return this;
	}
};

/*
struct D3D10ShaderResourceViewInPlace {
	D3D10_SHADER_RESOURCE_VIEW_DESC_CPP desc;

	D3D10ShaderResourceViewInPlace( DXGI_FORMAT Format ) : D3D10_SHADER_RESOURCE_VIEW_DESC_CPP( Format ) {}

	D3D10ShaderResourceViewInPlace & setTexture2D( UINT MostDetailedMip = 0, UINT MipLevels = 0 ) {
		desc.setTexture2D( MostDetailedMip, MipLevels );
		return *this;
	}

	D3D10_SHADER_RESOURCE_VIEW_DESC * operator D3D10_SHADER_RESOURCE_VIEW_DESC*() {
		return &desc;
	}

	static D3D10ShaderResourceViewInPlace texture2D( DXGI_FORMAT format, UINT MostDetailedMip = 0, UINT MipLevels = 0 ) {
		D3D10ShaderResourceViewInPlace inPlace( format );
		inPlace.setTexture2D( MostDetailedMip, MipLevels );
		return inPlace;
	}
 )
};*/

struct D3D10Texture2DDepthStencilViewDesc : D3D10_DEPTH_STENCIL_VIEW_DESC {
	D3D10Texture2DDepthStencilViewDesc( DXGI_FORMAT format ) {
		Format = format;
		ViewDimension = D3D10_DSV_DIMENSION_TEXTURE2D;
		Texture2D.MipSlice = 0;
	}
	
	D3D10Texture2DDepthStencilViewDesc &mipSlice( int level ) { Texture2D.MipSlice = level; return *this; }

	operator D3D10_DEPTH_STENCIL_VIEW_DESC*() {
		return this;
	}
};

struct D3D10Texture2DArrayDepthStencilViewDesc : D3D10_DEPTH_STENCIL_VIEW_DESC {
	D3D10Texture2DArrayDepthStencilViewDesc( DXGI_FORMAT format ) {
		Format = format;
		ViewDimension = D3D10_DSV_DIMENSION_TEXTURE2DARRAY;
		Texture2DArray.MipSlice = 0;
		Texture2DArray.FirstArraySlice = 0;
		Texture2DArray.ArraySize = 1;
	}

	D3D10Texture2DArrayDepthStencilViewDesc &mipSlice( int level ) { Texture2DArray.MipSlice = level; return *this; }
	D3D10Texture2DArrayDepthStencilViewDesc &firstArraySlice( int slice ) { Texture2DArray.FirstArraySlice = slice; return *this; }
	D3D10Texture2DArrayDepthStencilViewDesc &arraySize( int size ) { Texture2DArray.ArraySize = size; return *this; }

	operator D3D10_DEPTH_STENCIL_VIEW_DESC*() {
		return this;
	}
};

struct D3D10Texture2DRenderTargetViewDesc : D3D10_RENDER_TARGET_VIEW_DESC {
	D3D10Texture2DRenderTargetViewDesc( DXGI_FORMAT format ) {
		Format = format;
		ViewDimension = D3D10_RTV_DIMENSION_TEXTURE2D;
		Texture2D.MipSlice = 0;
	}

	D3D10Texture2DRenderTargetViewDesc & mipSlice( int level ) { Texture2D.MipSlice = level; return *this; }

	operator D3D10_RENDER_TARGET_VIEW_DESC*() {
		return this;
	}
};

struct D3D10Texture2DArrayRenderTargetViewDesc : D3D10_RENDER_TARGET_VIEW_DESC {
	D3D10Texture2DArrayRenderTargetViewDesc( DXGI_FORMAT format ) {
		Format = format;
		ViewDimension = D3D10_RTV_DIMENSION_TEXTURE2DARRAY;
		Texture2DArray.MipSlice = 0;
		Texture2DArray.FirstArraySlice = 0;
		Texture2DArray.ArraySize = 1;
	}

	D3D10Texture2DArrayRenderTargetViewDesc &mipSlice( int level ) { Texture2DArray.MipSlice = level; return *this; }
	D3D10Texture2DArrayRenderTargetViewDesc &firstArraySlice( int slice ) { Texture2DArray.FirstArraySlice = slice; return *this; }
	D3D10Texture2DArrayRenderTargetViewDesc &arraySize( int size ) { Texture2DArray.ArraySize = size; return *this; }

	operator D3D10_RENDER_TARGET_VIEW_DESC*() {
		return this;
	}
};

/*
template< typename T >
struct InPlace {
	T value;

	InPlace( const T &value ) : value( value ) {}

	operator T*() {
		return &value;
	}
};*/
