//--------------------------------------------------------------------------------------
// File: lpv.fx
//
// (FreeBSD License)
// 
// Copyright 2010 Andreas Kirsch. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification, are
// permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice, this list of
// conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice, this list
// of conditions and the following disclaimer in the documentation and/or other materials
// provided with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY ANDREAS KIRSCH ``AS IS'' AND ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
// FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ANDREAS KIRSCH OR
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
// ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// The views and conclusions contained in the software and documentation are those of the
// authors and should not be interpreted as representing official policies, either expressed
// or implied, of Andreas Kirsch.
//--------------------------------------------------------------------------------------

#ifdef _MSC_VER
#ifndef EFFECT_DEFINITION
#define EFFECT_NAME GPUInjection
#define EFFECT_FILE "lpv.fx"

#include "fxhelper.h"
#else
TECHNIQUE_( GpuInjection, gpuInjection )
TECHNIQUE_( GpuOcclusionInjection, gpuOcclusionInjection )
TECHNIQUE_( GpuPropagation, gpuPropagation )
TECHNIQUE_( GpuOccludedPropagation, gpuOccludedPropagation )

VARIABLE( ShaderResource, lpv )
VARIABLE( ShaderResource, gv )

VARIABLE( float, lpv_cellSize )
VARIABLE( D3DXVECTOR3, lpv_minCorner )
VARIABLE( D3DXVECTOR3, lpv_size )

VARIABLE( ShaderResource,  worldPosition )
VARIABLE( ShaderResource,  worldNormal )
VARIABLE( ShaderResource,  lightFlux )

#endif
#else

//const float Pi = 3.1415926f;
#define Pi 3.1415926f

//const float SH_c0 = 0.282094792f;
//const float SH_c1 = 0.488602512f;

#define SH_c0 0.282094792f // 1 / 2sqrt(pi)
#define SH_c1 0.488602512f // sqrt(3/pi) / 2

// helper function
float4 SH_evaluate( float3 direction ) {
	direction = normalize( direction );
	
	return float4( SH_c0, -SH_c1 * direction.y, SH_c1 * direction.z, -SH_c1 * direction.x );
}

// no normalization
float4 SH_evaluate_direct( float3 direction ) {	
	return float4( SH_c0, -SH_c1 * direction.y, SH_c1 * direction.z, -SH_c1 * direction.x );
}

// uses homogenous coordinate to become linear - and no normalization!
float4 SH_evaluate( float4 direction ) {	
	return dot( float4( SH_c0, -SH_c1, SH_c1, -SH_c1 ), direction.wyzx );
}


// my values from maple
#if 1
//const float SH_cosLobe_c0 = 0.886226925f; // sqrt(pi)/2
//const float SH_cosLobe_c1 = 1.02332671f; // sqrt(pi/3)
#define SH_cosLobe_c0 0.886226925f /* sqrt(pi)/2 */
#define SH_cosLobe_c1 1.02332671f /* sqrt(pi/3) */
#else
// the old crytek paper uses different values:
//const float SH_cosLobe_c0 = 0.25f;
//const float SH_cosLobe_c1 = 0.5f;
#define SH_cosLobe_c0 0.25f
#define SH_cosLobe_c1 0.5f
#endif

// helper function
float4 SH_evaluateCosineLobe( float3 direction ) {
	direction = normalize( direction );
	
	return float4( SH_cosLobe_c0, -SH_cosLobe_c1 * direction.y, SH_cosLobe_c1 * direction.z, -SH_cosLobe_c1 * direction.x );
}

// no normalization
float4 SH_evaluateCosineLobe_direct( float3 direction ) {
	return float4( SH_cosLobe_c0, -SH_cosLobe_c1 * direction.y, SH_cosLobe_c1 * direction.z, -SH_cosLobe_c1 * direction.x );
}

// uses homogenous coordinate to become linear - and no normalization!
float4 SH_evaluateCosineLobe_linear( float4 direction ) {
	return dot( float4( SH_cosLobe_c0, -SH_cosLobe_c1, SH_cosLobe_c1, -SH_cosLobe_c1 ), direction.wyzx );
}

DepthStencilState DepthState {
	DepthEnable = FALSE;
};

BlendState BlendAdditive
{
	BlendEnable[0] = true;
	BlendEnable[1] = true;
	BlendEnable[2] = true;
	SrcBlend = ONE;
	DestBlend = ONE;
	SrcBlendAlpha = ONE;
	DestBlendAlpha = ONE;
};

BlendState BlendPropagation
{
	BlendEnable[0] = false;
	BlendEnable[1] = false;
	BlendEnable[2] = false;
	BlendEnable[3] = true;
	BlendEnable[4] = true;
	BlendEnable[5] = true;
	SrcBlend = ONE;
	DestBlend = ONE;	
	SrcBlendAlpha = ONE;
	DestBlendAlpha = ONE;
};

SamplerState samOCV
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Border;
    AddressV = Border;
    AddressW = Border;
    BorderColor = float4( 0,0,0,0 );
};

// actual shader
Texture3D<float4> lpv[3];
Texture3D<float4> gv; // occlusion volume

float lpv_cellSize;
float3 lpv_minCorner;
int3 lpv_size;

// xyz store worldPos w stores lightPos.z
Texture2D<float4> worldPosition;
Texture2D<float3> worldNormal;
Texture2D<float3> lightFlux;

struct VS_INJECTION_INPUT {
	uint posIndex : SV_VertexID;
};

struct VS_INJECTION_OUTPUT {
	float4 cellIndex : SV_POSITION;
	float3 normal : WORLD_NORMAL;
	float3 flux : LIGHT_FLUX;
};

struct GS_INJECTION_OUTPUT {
	uint depthIndex : SV_RenderTargetArrayIndex;
	float4 screenPos : SV_Position;
	
	float3 normal : WORLD_NORMAL;
	float3 flux : LIGHT_FLUX;
};

struct PS_INJECTION_OUTPUT {
	float4 redSH : SV_Target0;
	float4 greenSH : SV_Target1;
	float4 blueSH : SV_Target2;
};

VS_INJECTION_OUTPUT VS_gpuInjection( VS_INJECTION_INPUT input ) {
	int2 RSMsize;
	worldPosition.GetDimensions( RSMsize.x, RSMsize.y );
	int3 rsmCoords = int3( input.posIndex % RSMsize.x, input.posIndex / RSMsize.x, 0 );
	
	float3 posWorld = worldPosition.Load( rsmCoords );
	float3 normalWorld = worldNormal.Load( rsmCoords );
	float3 flux = lightFlux.Load( rsmCoords );
	
	VS_INJECTION_OUTPUT output;
	output.cellIndex = float4( int3( (posWorld - lpv_minCorner) / lpv_cellSize + 0.5 * normalWorld ), 1.0 );
	
	output.normal = normalWorld;
	output.flux = flux;
	
	return output;
}

[maxvertexcount(1)]
void GS_gpuInjection( point VS_INJECTION_OUTPUT input[1], inout PointStream<GS_INJECTION_OUTPUT> OutputStream ) {
	GS_INJECTION_OUTPUT output;
	
	output.depthIndex = input[0].cellIndex.z;
	output.screenPos.xy = (float2(input[0].cellIndex.xy) + 0.5) / lpv_size.xy * 2.0 - 1.0;
	// invert y direction because y points downwards in the viewport?
	output.screenPos.y = -output.screenPos.y;
	output.screenPos.zw = float2( 0, 1 );

	output.normal = input[0].normal;
	output.flux = input[0].flux;
	
	OutputStream.Append( output );
}

PS_INJECTION_OUTPUT PS_gpuInjection( GS_INJECTION_OUTPUT input ) {
	PS_INJECTION_OUTPUT output;
	
	if( length( input.normal ) < 0.01 ) {
		discard;
	}
	
	
#if 0
	// this is what the paper says
	float4 SHcoeffs = SH_evaluateCosineLobe( input.normal );
	
	output.redSH = SHcoeffs * input.flux.r;
	output.greenSH = SHcoeffs * input.flux.g;
	output.blueSH = SHcoeffs * input.flux.b;
#else
	// the SHcoeffs model an intensity function
	// thus the integral over all intensities (over the sphere) should be the original flux
	// and that integral = Pi * totalFlux, so divide by Pi
	float4 SHcoeffs = SH_evaluateCosineLobe( input.normal ) / Pi;
	
	output.redSH = SHcoeffs * input.flux.r;
	output.greenSH = SHcoeffs * input.flux.g;
	output.blueSH = SHcoeffs * input.flux.b;
#endif
	
	return output;
}

struct VS_OCCLUSION_INJECTION_OUTPUT {
	float4 cellIndex : SV_POSITION;
	float3 normal : WORLD_NORMAL;
	float surfelArea : SURFEL_AREA;
};

struct GS_OCCLUSION_INJECTION_OUTPUT {
	uint depthIndex : SV_RenderTargetArrayIndex;
	float4 screenPos : SV_Position;
	
	float3 normal : WORLD_NORMAL;
	float surfelArea : SURFEL_AREA;
};

struct PS_OCCLUSION_INJECTION_OUTPUT {
	float4 occlusionSH : SV_Target0;
};

VS_OCCLUSION_INJECTION_OUTPUT VS_gpuOcclusionInjection( VS_INJECTION_INPUT input ) {
	int2 RSMsize;
	worldPosition.GetDimensions( RSMsize.x, RSMsize.y );
	int3 rsmCoords = int3( input.posIndex % RSMsize.x, input.posIndex / RSMsize.x, 0 );
	
	float4 posWorld = worldPosition.Load( rsmCoords );
	float3 normalWorld = worldNormal.Load( rsmCoords );
	
	VS_OCCLUSION_INJECTION_OUTPUT output;
	// shift the occlusion volume by half a cell
	output.cellIndex = float4( int3( (posWorld.xyz - lpv_minCorner) / lpv_cellSize - float3( 0.5, 0.5, 0.5 ) ), 1.0 );
	
	output.normal = normalWorld;
	
	output.surfelArea = 4.0 * posWorld.w * posWorld.w / RSMsize.x / RSMsize.y;
	
	return output;
}

[maxvertexcount(1)]
void GS_gpuOcclusionInjection( point VS_OCCLUSION_INJECTION_OUTPUT input[1], inout PointStream<GS_OCCLUSION_INJECTION_OUTPUT> OutputStream ) {
	GS_OCCLUSION_INJECTION_OUTPUT output;
	
	output.depthIndex = input[0].cellIndex.z;
	output.screenPos.xy = (float2(input[0].cellIndex.xy) + 0.5) / lpv_size.xy * 2.0 - 1.0;
	// invert y direction because y points downwards in the viewport?
	output.screenPos.y = -output.screenPos.y;
	output.screenPos.zw = float2( 0, 1 );

	output.normal = input[0].normal;
	output.surfelArea = input[0].surfelArea;
	
	OutputStream.Append( output );
}

//#define OCCINJ_OVERDRAW
//#define OCCINJ_TWOSIDED
PS_OCCLUSION_INJECTION_OUTPUT PS_gpuOcclusionInjection( GS_OCCLUSION_INJECTION_OUTPUT input ) {
	PS_OCCLUSION_INJECTION_OUTPUT output;
	
	if( length( input.normal ) < 0.01 ) {
		discard;
	}
	
	// this is what the paper says
#ifndef OCCINJ_TWOSIDED
	float4 SHcoeffs = SH_evaluateCosineLobe( input.normal );
#else
	float4 SHcoeffs = SH_evaluateCosineLobe_linear( float4( 0, 0, 0, 2 ) );
#endif
	
	// FIXME: just clamp the area ratio - if its bigger than > 1, it should actually be injected into more than just one voxel..
	float areaRatio = saturate( input.surfelArea / (lpv_cellSize *  lpv_cellSize) );
	
#ifdef OCCINJ_OVERDRAW
	int maxOverdraw = 1000;
	output.occlusionSH = float4( 1.0 / maxOverdraw, 0, 0, 0 );
#else
	output.occlusionSH = SHcoeffs * areaRatio;
#endif
	
	return output;
}

struct VS_PROP_INPUT {
	uint vertexIndex : SV_VertexID;
	uint depthIndex : SV_InstanceID;
};

struct VS_PROP_OUTPUT {
	float4 screenPos : SV_Position;
	uint depthIndex : DepthIndex;
};

struct GS_PROP_OUTPUT {
	float4 screenPos : SV_Position;
	uint depthIndex : SV_RenderTargetArrayIndex;
};

struct PS_PROP_OUTPUT {
	float4 redSH : SV_Target0;
	float4 greenSH : SV_Target1;
	float4 blueSH : SV_Target2;
	float4 acc_redSH: SV_Target3;
	float4 acc_greenSH : SV_Target4;
	float4 acc_blueSH : SV_Target5;
};

VS_PROP_OUTPUT VS_gpuPropagation( VS_PROP_INPUT input ) {
	VS_PROP_OUTPUT output;
	
	if( input.vertexIndex == 0 ) {
		output.screenPos.xy = float2( -1.0, 1.0 );
	}
	else if( input.vertexIndex == 1 ) {
		output.screenPos.xy = float2( 3.0, 1.0 );
	}
	else {
		output.screenPos.xy = float2( -1.0, -3.0 );
	}
	output.screenPos.zw = float2( 0.0, 1.0 );
	
	output.depthIndex = input.depthIndex;
	
	return output;
}

[maxvertexcount(3)]
void GS_gpuPropagation( triangle VS_PROP_OUTPUT input[3], inout TriangleStream<GS_PROP_OUTPUT> OutputStream ) {
	for( int i = 0 ; i < 3 ; i++ ) {
		GS_PROP_OUTPUT output;
		output.depthIndex = input[i].depthIndex;
		output.screenPos = input[i].screenPos;
		
		OutputStream.Append( output ); 
	}
}

struct SHcoeffs {
	float4 red, green, blue;
};

// right up
float2 Prop_side[4] = { float2( 1.0, 0.0 ), float2( 0.0, 1.0 ), float2( -1.0, 0.0 ), float2( 0.0, -1.0 ) };

// orientation = [ right | up | forward ] = [ x | y | z ]
float3 Prop_getEvalSideDirection( int index, float3x3 orientation ) {
	const float smallComponent = 0.4472135; // 1 / sqrt(5)
	const float bigComponent = 0.894427; // 2 / sqrt(5)
	
	const float2 side = Prop_side[ index ];
	// *either* x = 0 or y = 0
	return mul( orientation, float3( side.x * smallComponent, side.y * smallComponent, bigComponent ) );
}

float3 Prop_getReprojSideDirection( int index, float3x3 orientation ) {
	const float2 side = Prop_side[ index ];
	return mul( orientation, float3( side.x, side.y, 0 ) );
}

// orientation = [ right | up | forward ] = [ x | y | z ]
float3x3 Prop_neighbourOrientations[6] = {
	// Z+
	float3x3( 
		1, 0, 0,
		0, 1, 0,
		0, 0, 1
	),
	// Z-
	float3x3( 
		-1, 0, 0,
		0, 1, 0,
		0, 0, -1
	),
	// X+
	float3x3( 
		0, 0, 1,
		0, 1, 0,
		-1, 0, 0
	),
	// X-
	float3x3( 
		0, 0, -1,
		0, 1, 0,
		1, 0, 0
	),
	// Y+
	float3x3( 
		1, 0, 0,
		0, 0, 1,
		0, -1, 0
	),
	// Y-
	float3x3( 
		1, 0, 0,
		0, 0, -1,
		0, 1, 0
	)
};

#define GATHER_OCCLUSION
//#define GATHER_REFLECTION
SHcoeffs Prop_gatherContributions( int4 cellIndex, uniform bool sampleOcclusion ) {
	SHcoeffs contribution = { float4( 0.0, 0.0, 0.0, 0.0 ), float4( 0.0, 0.0, 0.0, 0.0 ), float4( 0.0, 0.0, 0.0, 0.0 ) };
	
	int3 OCVsize;
	gv.GetDimensions( OCVsize.x, OCVsize.y, OCVsize.z );

	for( int neighbour = 0 ; neighbour < 6 ; neighbour++ ) {
		float3x3 orientation = Prop_neighbourOrientations[ neighbour ];
		// TODO: transpose all orientation matrices and use row indexing instead? ie int3( orientation[2] )
		float3 mainDirection = mul( orientation, float3( 0, 0, 1 ) );
				
		int4 neighbourIndex = cellIndex - int4( mainDirection, 0 );
		SHcoeffs neighbourCoeffs = { lpv[0].Load( neighbourIndex ), lpv[1].Load( neighbourIndex ), lpv[2].Load( neighbourIndex ) };
		
	// crytek papers says that subtended solid angle has to be divided by 4 Pi, I disagree
#if 0
		const float directFaceSubtendedSolidAngle = 0.03188428; // 0.4006696846f / 4Pi;
		const float sideFaceSubtendedSolidAngle = 0.03369559; // 0.4234413544f / 4Pi;
#elif 0
		const float directFaceSubtendedSolidAngle = 4.0 / 18.0;
		const float sideFaceSubtendedSolidAngle = 4.0 / 36.0;
#else
		// again the contributions modelled below are approximating integrals over intensity distributions, ie total flux
		// we then backmodel the flux as a hemisphere (area) light, so to convert from flux to intensity we divide by Pi
		// and we do it here because its easier
		// TODO: since we're approximating with low order SH coeffs, maybe use a different/more precise *normalization* constant?
		const float directFaceSubtendedSolidAngle = 0.4006696846f / Pi;
		const float sideFaceSubtendedSolidAngle = 0.4234413544f / Pi;
#endif

		float transmittanceValue = 1.0; // no occlusion
		float reflectionValue = 0.0; // no reflection
#ifdef GATHER_OCCLUSION
		if( sampleOcclusion ) {
			float3 occCoord = (float3( neighbourIndex.xyz ) + 0.5 * mainDirection) / OCVsize;
			float4 occCoeffs = gv.Sample( samOCV, occCoord );
			// only evaluate one direction for now!
			transmittanceValue = 1.0 - saturate( occCoeffs * SH_evaluate_direct( -mainDirection ) );
			
#ifdef GATHER_REFLECTION
			float3 reflectionOccCoord = (float3( neighbourIndex.xyz ) + 1.5 * mainDirection) / OCVsize;
			float4 reflectionOccCoeffs = gv.Sample( samOCV, occCoord ); 
			// only reflect the light that passes through the cell -> weight with transmittance value..
			reflectionValue = transmittanceValue * saturate( reflectionOccCoeffs * SH_evaluate_direct( -mainDirection ) );
#endif
		}
#endif

		float occludedDirectFaceContribution = transmittanceValue * directFaceSubtendedSolidAngle;
		
		// TODO: simplify?
#ifndef GATHER_REFLECTION
		float4 mainDirectionCosineLobeSH = SH_evaluateCosineLobe_direct( mainDirection );
		float4 mainDirectionSH = SH_evaluate_direct( mainDirection );
		contribution.red += occludedDirectFaceContribution * dot( neighbourCoeffs.red, mainDirectionSH ) * mainDirectionCosineLobeSH;
		contribution.green += occludedDirectFaceContribution * dot( neighbourCoeffs.green, mainDirectionSH ) * mainDirectionCosineLobeSH;
		contribution.blue += occludedDirectFaceContribution * dot( neighbourCoeffs.blue, mainDirectionSH ) * mainDirectionCosineLobeSH;
#else
		float4 totalMainDirection = float4( mainDirection * (1.0 - reflectionValue), 1.0 + reflectionValue);
		
		float4 mainDirectionCosineLobeSH = SH_evaluateCosineLobe_linear( totalMainDirection );
		float4 mainDirectionSH = SH_evaluate_direct( mainDirection );
		
		contribution.red += occludedDirectFaceContribution * dot( neighbourCoeffs.red, mainDirectionSH ) * mainDirectionCosineLobeSH;
		contribution.green += occludedDirectFaceContribution * dot( neighbourCoeffs.green, mainDirectionSH ) * mainDirectionCosineLobeSH;
		contribution.blue += occludedDirectFaceContribution * dot( neighbourCoeffs.blue, mainDirectionSH ) * mainDirectionCosineLobeSH;
#endif
		for( int sideFace = 0 ; sideFace < 4 ; sideFace++ ) {
			float3 evalDirection = Prop_getEvalSideDirection( sideFace, orientation );
			float3 reprojDirection = Prop_getReprojSideDirection( sideFace, orientation );
			
#ifdef GATHER_OCCLUSION
			if( sampleOcclusion ) {
				float3 occCoord = (float3( neighbourIndex.xyz ) + 0.5 * evalDirection) / OCVsize;
				float4 occCoeffs = gv.Sample( samOCV, occCoord );
				// only evaluate one direction for now!
				transmittanceValue = 1.0 - saturate( occCoeffs * SH_evaluate_direct( -evalDirection ) );
			}
#endif
			
			float occludedSideFaceContribution = transmittanceValue * sideFaceSubtendedSolidAngle;
			
			float4 reprojDirectionCosineLobeSH = SH_evaluateCosineLobe_direct( reprojDirection );
			float4 evalDirectionSH = SH_evaluate_direct( evalDirection );
			
			contribution.red += occludedSideFaceContribution * dot( neighbourCoeffs.red, evalDirectionSH ) * reprojDirectionCosineLobeSH;
			contribution.green += occludedSideFaceContribution * dot( neighbourCoeffs.green, evalDirectionSH ) * reprojDirectionCosineLobeSH;
			contribution.blue += occludedSideFaceContribution * dot( neighbourCoeffs.blue, evalDirectionSH ) * reprojDirectionCosineLobeSH;
		}
	}
	
	return contribution;
}

PS_PROP_OUTPUT PS_gpuPropagation( float4 screenSpace : SV_Position, uint depthIndex : SV_RenderTargetArrayIndex, uniform bool sampleOcclusion ) {
	int4 cellIndex = int4( screenSpace.xy - 0.5f, depthIndex, 0 );
	
	PS_PROP_OUTPUT output;
	
	SHcoeffs contribution = Prop_gatherContributions( cellIndex, sampleOcclusion );
	
	// simply copy for now
	output.acc_redSH = output.redSH = contribution.red;
	output.acc_greenSH = output.greenSH = contribution.green;
	output.acc_blueSH = output.blueSH = contribution.blue;
	
	return output;
}

technique10 gpuInjection
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS_gpuInjection() ) );
        SetGeometryShader( CompileShader( gs_4_0, GS_gpuInjection() ) );
        SetPixelShader( CompileShader( ps_4_0, PS_gpuInjection() ) );
        
        SetDepthStencilState( DepthState, 0 );
        SetBlendState( BlendAdditive, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }
}

technique10 gpuOcclusionInjection
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS_gpuOcclusionInjection() ) );
        SetGeometryShader( CompileShader( gs_4_0, GS_gpuOcclusionInjection() ) );
        SetPixelShader( CompileShader( ps_4_0, PS_gpuOcclusionInjection() ) );
        
        SetDepthStencilState( DepthState, 0 );
        SetBlendState( BlendAdditive, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }
}

technique10 gpuPropagation
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS_gpuPropagation() ) );
        SetGeometryShader( CompileShader( gs_4_0, GS_gpuPropagation() ) );
        SetPixelShader( CompileShader( ps_4_0, PS_gpuPropagation( false ) ) );
        
        SetDepthStencilState( DepthState, 0 );
        SetBlendState( BlendPropagation, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }
}

technique10 gpuOccludedPropagation
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, VS_gpuPropagation() ) );
        SetGeometryShader( CompileShader( gs_4_0, GS_gpuPropagation() ) );
        SetPixelShader( CompileShader( ps_4_0, PS_gpuPropagation( true ) ) );
        
        SetDepthStencilState( DepthState, 0 );
        SetBlendState( BlendPropagation, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }
}

#endif


